from gidappdata import AppDataStorager
import pytest
import os
from gidappdata.utility.functions import readit, writeit, writebin, create_folder, writejson, pickleit, pathmaker, loadjson
import sys
import appdirs


def test_created_folder(simple_appdata_storage):
    assert simple_appdata_storage.appstorage_folder == pathmaker(appdirs.user_data_dir(appauthor='test_author', appname='test_app_name', roaming=True))
    assert simple_appdata_storage.log_folder == pathmaker(appdirs.user_log_dir(appauthor='test_author', appname='test_app_name', opinion=True))
    assert os.path.exists(simple_appdata_storage.appstorage_folder) is True
    assert os.path.exists(simple_appdata_storage.log_folder) is True
    assert os.listdir(simple_appdata_storage.log_folder) == []
    assert os.listdir(simple_appdata_storage.appstorage_folder) == []


@ pytest.mark.skipif(sys.platform != 'win32', reason="paths are specific to windows")
def test_folder_file_access(filled_appdata_storage):
    assert filled_appdata_storage['first_folder'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder"
    assert os.path.isdir(filled_appdata_storage['first_folder']) is True

    assert filled_appdata_storage['data_subfolder'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/data_subfolder"
    assert os.path.isdir(filled_appdata_storage['data_subfolder']) is True

    assert filled_appdata_storage['image_subfolder'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/image_subfolder"
    assert os.path.isdir(filled_appdata_storage['image_subfolder']) is True

    assert filled_appdata_storage['test_image.jpg'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/image_subfolder/test_image.jpg"
    assert os.path.isfile(filled_appdata_storage['test_image.jpg']) is True

    assert filled_appdata_storage['test_image.png'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/image_subfolder/test_image.png"
    assert os.path.isfile(filled_appdata_storage['test_image.png']) is True

    assert filled_appdata_storage['test_image.ico'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/image_subfolder/test_image.ico"
    assert os.path.isfile(filled_appdata_storage['test_image.ico']) is True

    assert filled_appdata_storage['test_1.txt'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/data_subfolder/test_1.txt"
    assert os.path.isfile(filled_appdata_storage['test_1.txt']) is True
    assert readit(filled_appdata_storage['test_1.txt']) == 'this is a test text file!'

    assert filled_appdata_storage['test_2.txt'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/data_subfolder/test_2.txt"
    assert os.path.isfile(filled_appdata_storage['test_2.txt']) is True
    assert readit(filled_appdata_storage['test_2.txt']) == 'this is another test text file!'

    assert filled_appdata_storage['test_dict.json'] == "C:/Users/Giddi/AppData/Roaming/test_author/test_app_name/first_folder/data_subfolder/test_dict.json"
    assert os.path.isfile(filled_appdata_storage['test_dict.json']) is True
    assert loadjson(filled_appdata_storage['test_dict.json']) == {}


def test_clear(filled_appdata_storage):
    main_folder = str(filled_appdata_storage).rsplit('/', 2)[0]

    assert os.path.isdir(main_folder) is True
    assert os.path.isdir(str(filled_appdata_storage)) is True
    assert os.path.isdir(filled_appdata_storage.log_folder) is True

    filled_appdata_storage.clean(filled_appdata_storage.AllFolder)

    assert os.path.isdir(str(filled_appdata_storage)) is False
    assert os.path.isdir(filled_appdata_storage.log_folder) is False
    assert os.path.isdir(main_folder) is True



# region [Imports]


import lzma
import os

import zipfile
import os
import base64
import logging
import gidlogger as glog
from gidconfig.standard import ConfigRental
from gidappdata.standard_appdata.appdata_storager import AppDataStorager
from gidappdata.utility.functions import pathmaker

# endregion [Imports]


# region [Logging]

log = logging.getLogger('gidappdata')
log.info(glog.imported(__name__))

# endregion [Logging]

# region [Constants]


# endregion [Constants]


# region [Configs]

def unzip(root_dir, zip_file, overwrite: bool = False):
    # sourcery skip: simplify-boolean-comparison
    with zipfile.ZipFile(zip_file, 'r') as zipf:
        for item in zipf.namelist():
            _info = zipf.getinfo(item)
            if _info.is_dir() is True:
                if os.path.isdir(pathmaker(root_dir, item)) is False:
                    os.makedirs(pathmaker(root_dir, item))
                    log.debug("created folder '%s' because it did not exist", pathmaker(root_dir, item))
                else:
                    log.debug("folder '%s' already exists", pathmaker(root_dir, item))
            else:
                if os.path.isfile(pathmaker(root_dir, item)) is False:
                    zipf.extract(item, pathmaker(root_dir))
                    log.debug("extracted file '%s' because it didn't exist", pathmaker(root_dir, item))
                elif overwrite is True:
                    log.debug("file '%s' already exist and is overwriten because overwrite is 'True'", pathmaker(root_dir, item))
                    zipf.extract(item, pathmaker(root_dir))
                else:
                    log.debug("file '%s' is already existing and overwrite is 'False' so file was not extracted", pathmaker(root_dir, item))


# endregion [Configs]


# region [Factories]

class AppdataProvider:
    handler = None

    @classmethod
    def setup_appdata(cls, author_name: str, app_name: str, folderlist: list = None, filelist: list = None, configs: dict = None, dev=None, redirect=None):
        if cls.handler is None:
            cls.handler = AppDataStorager(author_name, app_name, dev, redirect)
        if folderlist is not None:
            for _item in folderlist:
                if isinstance(_item, str):
                    cls.handler.add_folder(_item)
                elif isinstance(_item, tuple):
                    cls.handler.add_folder(_item[0], _item[1])
        if filelist is not None:
            for _item in filelist:
                if _item[0].endswith('.json'):
                    cls.handler.write_json(*_item)
                else:
                    cls.handler.write(*_item)

        if configs is not None:

            cls.handler.generate_configs(**configs)
        ConfigRental.set_appdata(cls.handler)
        return cls.handler

    @classmethod
    def archive_from_bin(cls, bin_data: str, name: str = 'user_data_archive', ext: str = 'zip', uses_base64: bool = False):
        _file = pathmaker(str(cls.handler), name + '.' + ext)
        with open(_file, 'wb') as archfile:
            if uses_base64 is True:
                bin_data = base64.b64decode(bin_data)
            archfile.write(bin_data)
        return _file

    @classmethod
    def unpack_archive(cls, in_archive_file, clean: bool):
        unzip(str(cls.handler), in_archive_file, False)
        if clean:
            os.remove(in_archive_file)

    @classmethod
    def setup_from_binarchive(cls, author_name: str, app_name: str, in_archive: str, uses_base64: bool, dev=None, redirect=None, clean=True):

        if cls.handler is None:
            log.info("appdata, does not exist, creating from scratch")
            cls.handler = AppDataStorager(author_name, app_name, dev, redirect)
            _archive = cls.archive_from_bin(in_archive, uses_base64=uses_base64)
            cls.unpack_archive(_archive, clean=clean)
            ConfigRental.set_appdata(cls.handler)
        else:
            log.info("appdata, already existing so returning existing object")

        return cls.handler

    @classmethod
    def get_handler(cls):
        if cls.handler is not None:
            return cls.handler
        else:
            raise LookupError('AppDataStorage object has to be created first via "get_handler" of this factory')

# endregion [Factories]


# region [Main_Exec]
if __name__ == '__main__':
    pass


# endregion [Main_Exec]



class GidAppDataBaseException(Exception):
    pass


class ConstructionEnvDataMissing(GidAppDataBaseException):
    def __init__(self, env_var_name):
        self.message = f"Variable '{env_var_name}' has not been set in the construction env file"
        super().__init__(self.message)


class DevSettingError(GidAppDataBaseException):
    def __init__(self):
        self.message = "'redirect' has to be an directory if 'dev' is set to 'True'"
        super().__init__(self.message)


class IsDuplicateNameError(GidAppDataBaseException):
    def __init__(self, name: str, full_path, is_file: bool = True):
        self.name = name
        self.full_path = full_path
        self.typus = "file" if is_file is True else "folder"
        self.msg = f"The {self.typus} name {self.name} ('{self.full_path}') already exists in this user_data dir, all file names and folder names need to be unique."
        super().__init__(self.msg)



# region [Imports]


from pprint import pformat
from enum import Enum, auto
import os
import sys
import shutil
import logging
from gidappdata.utility.functions import appendwriteit, linereadit, pathmaker, readit, writeit, writejson, loadjson, create_folder

import gidlogger as glog

import appdirs

# endregion [Imports]


# region [Logging]

log = logging.getLogger('gidappdata')
log.info(glog.imported(__name__))

# endregion [Logging]


# region [Constants]


# endregion [Constants]


class _AppDataFolder(Enum):
    AllFolder = auto()
    LogFolder = "log_folder"
    AppStorageFolder = "appstorage_folder"


class AppDataStorager:
    # region [ClassAttributes]

    AllFolder = _AppDataFolder.AllFolder
    LogFolder = _AppDataFolder.LogFolder
    AppStorageFolder = _AppDataFolder.AppStorageFolder

# endregion[ClassAttributes]
# region [Init]

    def __init__(self, author_name: str, app_name: str, dev: bool = False, redirect: str = None, log_folder: str = None):
        # sourcery skip: simplify-boolean-comparison
        self.dev = dev
        self.author_name = author_name
        self.app_name = app_name
        self.redirect = redirect
        self.managed_folder = []
        self.operating_system = sys.platform
        self.appstorage_folder = None if self.dev is False else self.redirect
        self.log_folder = None if self.dev is False else pathmaker(log_folder, 'logs')
        if self.dev is False:
            self.setup_app_storage_base()

# endregion[Init]
# region [Properties]

    @property
    def folder(self):
        _out = {}
        for dirname, dirlist, _ in os.walk(self.appstorage_folder):
            for _dir in dirlist:
                _out[_dir] = pathmaker(dirname, _dir)
        return _out

    @property
    def files(self):
        _out = {}
        for dirname, _, filelist in os.walk(self.appstorage_folder):
            for _file in filelist:
                _out[_file] = pathmaker(dirname, _file)
        return _out

# endregion[Properties]

    def setup_app_storage_base(self):
        self.appstorage_folder = pathmaker(appdirs.user_data_dir(appauthor=self.author_name, appname=self.app_name, roaming=True))
        self.managed_folder.append(self.appstorage_folder)
        if os.path.isdir(self.appstorage_folder) is False:
            os.makedirs(self.appstorage_folder)

        self.log_folder = pathmaker(appdirs.user_log_dir(appauthor=self.author_name, appname=self.app_name, opinion=True))
        self.managed_folder.append(self.log_folder)
        if os.path.isdir(self.log_folder) is False:
            os.makedirs(self.log_folder)

    def add_folder(self, folder_name, parent_folder=None):
        if parent_folder is None:
            _folder = pathmaker(self.appstorage_folder, folder_name)
        else:
            _folder = pathmaker(self[parent_folder], folder_name)
        if os.path.isdir(_folder) is False:
            os.makedirs(_folder)

    def copy_file(self, source, target_filename, folder=None, overwrite=False):
        _path = self._get_filepath(target_filename, folder)
        if os.path.isfile(_path) is False or overwrite is True:
            shutil.copyfile(source, _path)

    def _get_filepath(self, filename, folder):
        if folder is not None:
            return pathmaker(self.folder[folder], filename)
        else:
            return pathmaker(self.appstorage_folder, filename)

    def _get_app_base_folder(self, in_folder):
        _folder = in_folder
        while os.path.basename(_folder) != self.author_name:
            _folder = _folder.rsplit('/', 1)[0]
        return _folder

    def clean(self, folder_to_clean: _AppDataFolder):
        _to_delete = []
        if folder_to_clean is self.AllFolder:
            _to_delete = self.managed_folder
        else:
            _to_delete.append(getattr(self, folder_to_clean.value))

        for _folder in _to_delete:
            _base_folder = self._get_app_base_folder(_folder)
            shutil.rmtree(_base_folder)
            log.info('deleted appdata folder "%s"', _base_folder)

# region [SpecialMethods]

    def __getitem__(self, key):
        _out = None
        if key in self.files:
            _out = self.files[key]
        elif key in self.folder:
            _out = self.folder[key]
        else:
            if '.' in key:
                create_folder(pathmaker(self.appstorage_folder, 'unfoldered_files'))
                _out = pathmaker(self.appstorage_folder, 'unfoldered_files', key)
                log.debug("file '%s' does not exist, providing path to 'unfoldered_files' folder", key)
            else:
                log.debug("folder '%s' does not exist, providing path to hypothetical folder in base dir", key)
                _out = pathmaker(self.appstorage_folder, key)
        return _out

    def __str__(self):
        return self.appstorage_folder

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.author_name}, {self.app_name}, {str(self.dev)}, {str(self.redirect)})"

# endregion[SpecialMethods]



# taskarg: ${fileDirname}

import os
import json
import shutil
import base64
from dotenv import load_dotenv
from gidappdata.utility.functions import pathmaker, writebin, writeit, writejson, readbin, readit
from gidappdata.utility.exceptions import IsDuplicateNameError
from time import sleep, time
import gidlogger as glog
import click
import logging
import glob
from tempfile import TemporaryDirectory
import checksumdir

log = logging.getLogger('gidappdata')

log.info(glog.NEWRUN())


def as_kb(in_size: int) -> int:
    """
    Converter functions to convert the size in bytes to kilobytes.
    """
    conv = 1024
    return in_size // conv


def as_mb(in_size: int) -> int:
    """
    Converter functions to convert the size in bytes to megabytes.
    """
    conv = 1024 * 1024
    return in_size // conv


def as_gb(in_size: int) -> int:
    """
    Converter functions to convert the size in bytes to gigabytes.
    """
    conv = 1024 * 1024 * 1024
    return in_size // conv


def find_files():
    # sourcery skip: inline-immediately-returned-variable, list-comprehension
    _out = []
    for _file in os.scandir():
        if not _file.name.endswith('.py') and not os.path.isdir(_file.path):
            _out.append(_file.path)
    return _out


def pack_data(in_path):
    _folder = pathmaker(in_path, 'data_pack')
    archive_path = pathmaker(shutil.make_archive(pathmaker(in_path, 'base_userdata_archive'), format='zip', root_dir=_folder, logger=log))
    log.info('data was archived with size of %s%s', *convert_file_size(archive_path))
    return archive_path


def convert_to_bin(archive, use_base64):
    with open(archive, 'rb') as binf:
        _content = binf.read()
    if use_base64 is True:
        _content = base64.b64encode(_content)
    return _content


def write_to_pyfile(in_path, **kwargs):
    _path = pathmaker(in_path, 'bin_data.py')
    with open(_path, 'w') as _file:
        for key, value in kwargs.items():
            _content = value
            _file.write(f'{key.strip()} = {_content.strip()}\n\n')
    log.info("bin data was written to python file: '%s'", _path)
    return _path


def write_construction_info(in_path, appname, author='BrocaProgs', uses_base64=True):
    _path = pathmaker(in_path, 'construction_info.env')
    with open(_path, 'w') as confo_file:
        confo_file.write(f"USES_BASE64 = {str(uses_base64)}\n")
        confo_file.write(f"AUTHOR_NAME = {str(author)}\n")
        confo_file.write(f"APP_NAME = {str(appname)}\n")

    log.info("construction info file was written to python file: '%s'", _path)
    return _path


def convert_file_size(in_file_path):
    size = os.stat(in_file_path).st_size
    if as_gb(size) > 1:
        log_size = round(as_gb(size), 3)
        log_size_type = 'gb'
    elif as_mb(size) > 1:
        log_size = round(as_mb(size), 3)
        log_size_type = 'mb'
    elif as_kb(size) > 1:
        log_size = round(as_kb(size), 1)
        log_size_type = 'kb'
    else:
        log_size = size
        log_size_type = 'b'
    return log_size, log_size_type


def post_clean_zip_file(file_path):
    log.info("starting cleanup of zip file")
    os.remove(file_path)
    if os.path.isfile(file_path) is False:
        log.info("cleanup of zip file successfully done")
    else:
        log.critical("was not able to remove zip file '%s'", file_path)


def read_const_file(const_file):
    _out = []
    with open(const_file, 'r') as c_file:
        content = c_file.read().splitlines()
    for line in content:
        if line != '' and ' = ' in line:
            _name, _value = line.split(' = ')
            _out.append((_name, _value))
    return _out


def post_checks(bin_py_file, const_file):

    log.info('checking existance of binary pyfile:')

    if os.path.exists(bin_py_file) is True:
        log.info("binary pyfile does exist at '%s'", bin_py_file)
        log.info("--> size: %s%s", *convert_file_size(bin_py_file))
    else:
        log.critical("binary pyfile does NOT exist! should be at '%s'", bin_py_file)

    log.info('checking existance of construction info file:')
    if os.path.exists(const_file) is True:
        log.info("construction info file does exist at '%s'", const_file)
        log.info("set variables found in construction info file: ")
        log.info("-----------------------")
        for name, value in read_const_file(const_file):
            log.info("%s = '%s'", name, value)
            log.info("-----------------------")
    else:
        log.critical("construction info file does NOT exist! should be at '%s'", const_file)


def get_ignore_pattern_files(base_folder, pattern, file_list):
    for file in glob.iglob(base_folder + '/**/' + pattern, recursive=True):
        file_list.append(pathmaker(os.path.abspath(file)))


def temporary_move_out(filelist, tempdir):
    _out = []
    for file in filelist:
        print('moving out "%s"', os.path.basename(file))
        new_path = pathmaker(tempdir, os.path.basename(file))

        _out.append((file, new_path))
        shutil.move(file, new_path)
    return _out


def move_back(filelist):
    for file in filelist:
        shutil.move(file[1], file[0])


def duplicate_check(in_dir):
    collected_folder = []
    collected_files = []
    for dirname, folderlist, filelist in os.walk(in_dir):
        for folder in folderlist:
            if folder not in collected_folder:
                collected_folder.append(folder)
            else:
                raise IsDuplicateNameError(folder, pathmaker(dirname, folder), False)
        for file in filelist:
            if file not in collected_files:
                collected_files.append(file)
            else:
                raise IsDuplicateNameError(file, pathmaker(dirname, file), True)


@click.command()
@click.argument('init_userdata_dir')
@click.option('-n', '--appname', default=os.getenv('PROJECT_NAME'))
@click.option('-a', '--author', default='BrocaProgs')
@click.option('--use-base64/--dont-base64', '-64/-no64', default=True)
@click.option('--clean-zip-file/--keep-zip-file', '-cz/-kz', default=True)
@click.option('--ignore-pattern', '-i', multiple=True)
def cli_generate_user_data_binfile(init_userdata_dir, appname, author, use_base64, clean_zip_file, ignore_pattern):

    start_time = time()
    duplicate_check(init_userdata_dir)
    tempdir = TemporaryDirectory()
    tempdir_name = tempdir.name
    ignore_files = []
    for pattern in ignore_pattern:
        get_ignore_pattern_files(init_userdata_dir, pattern, ignore_files)
    moved_files = temporary_move_out(ignore_files, tempdir_name)
    if appname is None or appname == '':
        print('Unable to obtain "appname" from env variable, please set "PROJECT_NAME" env variable or provide appname as cli-option')
        return
    appname = appname.replace(' ', '-').replace('_', '-').title()
    author = author.replace(' ', '-').replace('_', '-').title()
    log.info("Starting conversion for data_pack in '%s'")

    _archive = pack_data(init_userdata_dir)

    log.info('converted archive to bin')

    _py_file = write_to_pyfile(init_userdata_dir, bin_archive_data=convert_to_bin(_archive, use_base64))

    _const_file = write_construction_info(in_path=init_userdata_dir, appname=appname, author=author, uses_base64=use_base64)

    if clean_zip_file is True:
        post_clean_zip_file(_archive)

    log.info('running post-checks')
    post_checks(_py_file, _const_file)
    move_back(moved_files)
    tempdir.cleanup()

    log.debug('overall time taken: %s seconds', str(round(time() - start_time, 3)))
    log.info('---done---')


def generate_user_data_binfile(init_userdata_dir, appname, author, use_base64=True, clean_zip_file=True, ignore_pattern=None):
    start_time = time()
    duplicate_check(init_userdata_dir)
    tempdir = TemporaryDirectory()
    tempdir_name = tempdir.name
    ignore_files = []
    for pattern in ignore_pattern:
        get_ignore_pattern_files(init_userdata_dir, pattern, ignore_files)
    moved_files = temporary_move_out(ignore_files, tempdir_name)
    if appname is None or appname == '':
        print('Unable to obtain "appname" from env variable, please set "PROJECT_NAME" env variable or provide appname as cli-option')
        return
    appname = appname.replace(' ', '-').replace('_', '-').title()
    log.info("Starting conversion for data_pack in '%s'")
    checksum_userdata = checksumdir.dirhash(pathmaker(init_userdata_dir, 'data_pack'))
    _archive = pack_data(init_userdata_dir)

    log.info('converted archive to bin')
    bin_data = convert_to_bin(_archive, use_base64)
    _py_file = write_to_pyfile(init_userdata_dir, bin_archive_data=bin_data)

    _const_file = write_construction_info(in_path=init_userdata_dir, in_hash=checksum_userdata, appname=appname, author=author, uses_base64=use_base64)

    if clean_zip_file is True:
        post_clean_zip_file(_archive)

    log.info('running post-checks')
    post_checks(_py_file, _const_file)
    move_back(moved_files)
    tempdir.cleanup()
    log.debug('overall time taken: %s seconds', str(round(time() - start_time, 3)))
    log.info('---done---')
    return bin_data


if __name__ == '__main__':
    # THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))
    # _archive = pack_data(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppData\gidappdata\data\skeletons\prebuilt_standard\basic")
    # _py_file = write_to_pyfile(THIS_FILE_DIR, bin_archive_data=convert_to_bin(_archive, True))
    pass



import pytest
from gidappdata.cli.skeleton_tree import SkeletonInstructionItem, SkeletonTypus, DirSkeletonReader
from gidappdata.utility.functions import readbin, readit, pathmaker, loadjson
import os

THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))

RESTRUCTURE_REASON = "implementing new child Item method soon"


def helper_child_walker(tree, name):
    for child in tree.children:
        if child.name == name:
            return child


def test_child_access(access_test_tree):
    root, first_folder_item, second_folder_item, first_sub_folder_item, sub_folder_file_item = access_test_tree
    assert root.first_folder is first_folder_item
    assert root.first_folder.first_sub_folder is first_sub_folder_item
    assert root.second_folder is second_folder_item
    with pytest.raises(AttributeError):
        assert root.non_existing_file__ini
    assert root.first_folder.first_sub_folder.sub_folder_file__txt is sub_folder_file_item


def test_paths(simple_directory_tree, temp_root_dir):
    assert set(simple_directory_tree.get_paths()) == set(['root',
                                                          'root/first_folder',
                                                          'root/first_folder/first_folder_image_file.jpg',
                                                          'root/first_folder/first_folder_ini_file.ini',
                                                          'root/first_folder/first_subfolder',
                                                          'root/first_folder/second_subfolder',
                                                          'root/second_folder',
                                                          'root/text_file.txt',
                                                          'root/first_folder/first_subfolder/nested_test.json',
                                                          'root/second_folder/something_file.txt'])
    simple_directory_tree.set_root_path(temp_root_dir)
    assert set(simple_directory_tree.get_paths()) == set([pathmaker(temp_root_dir),
                                                          pathmaker(temp_root_dir, 'first_folder'),
                                                          pathmaker(temp_root_dir, 'first_folder/first_folder_image_file.jpg'),
                                                          pathmaker(temp_root_dir, 'first_folder/first_folder_ini_file.ini'),
                                                          pathmaker(temp_root_dir, 'first_folder/first_subfolder'),
                                                          pathmaker(temp_root_dir, 'first_folder/second_subfolder'),
                                                          pathmaker(temp_root_dir, 'second_folder'),
                                                          pathmaker(temp_root_dir, 'text_file.txt'),
                                                          pathmaker(temp_root_dir, 'first_folder/first_subfolder/nested_test.json'),
                                                          pathmaker(temp_root_dir, 'second_folder', 'something_file.txt')])


def test_file_children(simple_directory_tree):
    with pytest.raises(AttributeError):
        simple_directory_tree.text_file__txt.add_child_item(SkeletonInstructionItem('this_raises_Error', SkeletonTypus.Folder))
    with pytest.raises(AttributeError):
        simple_directory_tree.text_file__txt.add_child_item(SkeletonInstructionItem('this_raises_Error.csv', SkeletonTypus.File, content='this,should,raise,an,AttributeError'))


def test_existing_children(simple_directory_tree):
    with pytest.raises(FileExistsError):
        simple_directory_tree.add_child_item(SkeletonInstructionItem('text_file.txt', SkeletonTypus.File, content='this should raise an FileExistsError'))


def test_root_instantiation():
    root = SkeletonInstructionItem('root', SkeletonTypus.Root)
    assert root.name == 'root'
    assert root.typus is SkeletonTypus.Root
    assert root.content is None
    assert root.parent is None
    assert root.children == []


def test_dynamic_path_attribute(temp_root_dir):
    root = SkeletonInstructionItem('root', SkeletonTypus.Root)
    folder_level_0 = SkeletonInstructionItem('0_level_folder', SkeletonTypus.Folder)
    folder_level_1 = SkeletonInstructionItem('1_level_folder', SkeletonTypus.Folder)
    file_level_1_1 = SkeletonInstructionItem('1_1_level_file', SkeletonTypus.File, content='Bla Bla Bla')
    root.add_child_item(folder_level_0)
    folder_level_0.add_child_item(folder_level_1)
    folder_level_0.add_child_item(file_level_1_1)
    assert root.path == pathmaker('root')
    assert folder_level_0.path == pathmaker('root', '0_level_folder')
    assert folder_level_1.path == pathmaker('root', '0_level_folder', '1_level_folder')
    assert file_level_1_1.path == pathmaker('root', '0_level_folder', '1_1_level_file')
    root.set_root_path(temp_root_dir)
    assert root.path == pathmaker(temp_root_dir)
    assert folder_level_0.path == pathmaker(temp_root_dir, '0_level_folder')
    assert folder_level_1.path == pathmaker(temp_root_dir, '0_level_folder', '1_level_folder')
    assert file_level_1_1.path == pathmaker(temp_root_dir, '0_level_folder', '1_1_level_file')


def test_set_root_path(simple_directory_tree, temp_root_dir):
    simple_directory_tree.set_root_path(temp_root_dir)
    assert simple_directory_tree.name == temp_root_dir
    assert simple_directory_tree.typus is SkeletonTypus.Root
    assert simple_directory_tree.path == pathmaker(temp_root_dir)
    first_folder = simple_directory_tree.first_folder
    assert first_folder.name == 'first_folder'
    assert first_folder.typus is SkeletonTypus.Folder
    assert first_folder.path == pathmaker(temp_root_dir, 'first_folder')
    subfolder = first_folder.first_subfolder
    assert subfolder.name == 'first_subfolder'
    assert subfolder.typus is SkeletonTypus.Folder
    assert subfolder.path == pathmaker(temp_root_dir, 'first_folder', 'first_subfolder')


def test_file_handling(simple_directory_tree):
    first_file = simple_directory_tree.text_file__txt
    assert first_file.children is None
    assert first_file.typus is SkeletonTypus.File


def test_build_correct_names(simple_directory_tree, temp_root_dir):
    simple_directory_tree.set_root_path(temp_root_dir)
    simple_directory_tree.start_build()
    assert set(os.listdir(temp_root_dir)) == set(['text_file.txt', 'second_folder', 'first_folder'])
    assert set(os.listdir(pathmaker(temp_root_dir, 'first_folder'))) == set(['first_subfolder', 'second_subfolder', 'first_folder_image_file.jpg', 'first_folder_ini_file.ini'])


def test_correct_file_content(simple_directory_tree, temp_root_dir):
    simple_directory_tree.set_root_path(temp_root_dir)
    simple_directory_tree.start_build()
    with open(pathmaker(temp_root_dir, 'text_file.txt'), 'r') as fp:
        assert fp.read() == 'this is a test text'
    with open(pathmaker(temp_root_dir, 'first_folder', 'first_folder_ini_file.ini'), 'r') as fpini:
        assert fpini.read() == readit(pathmaker('tests', 'skeleton_tree_test', 'example_cfg.ini'))
    with open(pathmaker(temp_root_dir, 'first_folder', 'first_folder_image_file.jpg'), 'rb') as fpb:
        assert fpb.read() == readbin(pathmaker('tests', 'skeleton_tree_test', "afa_logoover_ca.jpg"))


def test_find_node(simple_directory_tree):
    x = simple_directory_tree.find_node('second_subfolder')
    assert x.name == 'second_subfolder'
    assert x.typus == SkeletonTypus.Folder
    assert x.content is None
    y = simple_directory_tree.find_node('nested_test.json')
    assert y.name == 'nested_test.json'
    assert y.typus == SkeletonTypus.File
    assert y.content == '{}'


def test_serialize(tmpdir):
    source_dir = pathmaker(THIS_FILE_DIR, 'example_dir', 'data_pack')
    target = tmpdir.join('serialize_test.json')
    tree = DirSkeletonReader(source_dir)
    tree.serialize(target, strategy='json')
    assert loadjson(target) == loadjson(pathmaker(THIS_FILE_DIR, 'serialize_test_compare.json'))


def test_build_from_json(tmpdir):
    target = tmpdir.mkdir('test_tree_build')
    tree = SkeletonInstructionItem.from_json_file(pathmaker(THIS_FILE_DIR, 'serialize_test_compare.json'))
    tree.set_root_path(target)
    tree.start_build()
    for path in tree.get_paths():
        assert os.path.exists(path) is True



# region [Imports]


# * Standard Library Imports -->
import os
import json
import shutil
import pickle
# * Gid Imports -->
import gidlogger as glog
from functools import partial
import base64
# import numpy as np
# endregion[Imports]


# region [Logging]

log = glog.logging.getLogger('gidappdata')
log.info(glog.imported(__name__))

# endregion[Logging]

# region [Constants]

# endregion[Constants]


def create_folder(in_folder, overwrite=False):
    if os.path.isdir(in_folder) is True:
        if overwrite is True:
            shutil.rmtree(in_folder)
        elif overwrite is False:
            log.debug('folder "%s" already exists', in_folder)
    if os.path.isdir(in_folder) is False:
        os.makedirs(in_folder)
        log.debug('created folder "%s"', in_folder)


def create_file(in_file, file_content, overwrite=False):
    if os.path.isfile(in_file) is False or overwrite is True:
        if os.path.splitext(in_file)[1] == '.json':
            writejson(file_content, in_file, sort_keys=False, indent=2)
        else:
            _write_mode = 'wb' if isinstance(file_content, bytes) else 'w'
            content = file_content if isinstance(file_content, bytes) else str(file_content)
            with open(in_file, _write_mode) as outfile:
                outfile.write(content)
        log.debug('created file "%s"', in_file)
    else:
        log.debug('file "%s" already exists', in_file)


def pathmaker(first_segment, *in_path_segments, rev=False):
    """
    Normalizes input path or path fragments, replaces '\\\\' with '/' and combines fragments.

    Parameters
    ----------
    first_segment : str
        first path segment, if it is 'cwd' gets replaced by 'os.getcwd()'
    rev : bool, optional
        If 'True' reverts path back to Windows default, by default None

    Returns
    -------
    str
        New path from segments and normalized.
    """
    _first = os.getcwd() if first_segment == 'cwd' else first_segment
    _path = os.path.join(_first, *in_path_segments)
    _path = _path.replace('\\\\', '/')
    _path = _path.replace('\\', '/')
    if rev is True:
        _path = _path.replace('/', '\\')

    return _path.strip()


def writebin(in_file, in_data):
    """
    Writes a string to binary.

    Parameters
    ----------
    in_file : str
        The target file path
    in_data : str
        The data to write
    """
    with open(in_file, 'wb') as outbinfile:
        outbinfile.write(in_data)


def writeit(in_file, in_data, append=False, in_encoding='utf-8', in_errors=None):
    """
    Writes to a file.

    Parameters
    ----------
    in_file : str
        The target file path
    in_data : str
        The data to write
    append : bool, optional
        If True appends the data to the file, by default False
    in_encoding : str, optional
        Sets the encoding, by default 'utf-8'
    """
    _write_type = 'w' if append is False else 'a'
    with open(in_file, _write_type, encoding=in_encoding, errors=in_errors,) as _wfile:
        _wfile.write(in_data)


def appendwriteit(in_file, in_data, in_encoding='utf-8'):
    with open(in_file, 'a', encoding=in_encoding) as appendwrite_file:
        appendwrite_file.write(in_data)


def readbin(in_file):
    """
    Reads a binary file.

    Parameters
    ----------
    in_file : str
        A file path

    Returns
    -------
    str
        the decoded file as string
    """
    with open(pathmaker(in_file), 'rb') as binaryfile:
        return binaryfile.read()


def readit(in_file, per_lines=False, in_encoding='utf-8', in_errors='replace'):
    """
    Reads a file.

    Parameters
    ----------
    in_file : str
        A file path
    per_lines : bool, optional
        If True, returns a list of all lines, by default False
    in_encoding : str, optional
        Sets the encoding, by default 'utf-8'
    in_errors : str, optional
        How to handle encoding errors, either 'strict' or 'ignore', by default 'strict'

    Returns
    -------
    str/list
        the read in file as string or list (if per_lines is True)
    """
    with open(in_file, 'r', encoding=in_encoding, errors=in_errors) as _rfile:
        _content = _rfile.read()
    if per_lines is True:
        _content = _content.splitlines()

    return _content


def linereadit(in_file, in_encoding='utf-8', in_errors='strict'):
    with open(in_file, 'r', encoding=in_encoding, errors=in_errors) as lineread_file:
        _out = lineread_file.read().splitlines()
    return _out


def clearit(in_file):
    """
    Deletes the contents of a file.

    Parameters
    ----------
    in_file : str
        The target file path
    """
    with open(in_file, 'w') as file_to_clear:
        file_to_clear.write('')
    log.debug(f"contents of file '{in_file}' was cleared")


def loadjson(in_file):
    with open(in_file, 'r') as jsonfile:
        _out = json.load(jsonfile)
    return _out


def writejson(in_object, in_file, sort_keys=True, indent=0):
    with open(in_file, 'w') as jsonoutfile:
        json.dump(in_object, jsonoutfile, sort_keys=sort_keys, indent=indent)


def pickleit(obj, in_path):
    """
    saves an object as pickle file.

    Parameters
    ----------
    obj : object
        the object to save
    in_name : str
        the name to use for the pickled file
    in_dir : str
        the path to the directory to use
    """
    with open(pathmaker(in_path), 'wb') as filetopickle:
        log.debug(f"saved object [{str(obj)}] as pickle file [{in_path}]")
        pickle.dump(obj, filetopickle, pickle.HIGHEST_PROTOCOL)


def get_pickled(in_path):
    """
    loads a pickled file.

    Parameters
    ----------
    in_path : str
        the file path to the pickle file

    Returns
    -------
    object
        the pickled object
    """
    with open(pathmaker(in_path), 'rb') as pickletoretrieve:
        log.debug(f"loaded pickle file [{in_path}]")
        return pickle.load(pickletoretrieve)


# def np_readbin(in_path):
#     return np.frombuffer(readbin(in_path))


def read_file(in_file):
    _read_strategies = {'.txt': readit,
                        '.ini': readit,
                        '.json': loadjson,
                        '.jpg': readbin,
                        '.png': readbin,
                        '.tga': readbin,
                        '.ico': readbin,
                        '.pkl': readbin,
                        '.py': readit,
                        '.cmd': readit,
                        '.env': readit,
                        '.log': readit,
                        '.errors': readit,
                        '.bat': readit,
                        '.md': readit}
    _ext = os.path.splitext(in_file)[1]
    return _read_strategies.get(_ext, readbin)(in_file)


def to_attr_name(in_name):

    replace_dict = {' ': '_',
                    '-': '_',
                    '.': '__',
                    '/': '_',
                    '\\': '_',
                    '*': '',
                    '{': '_',
                    '}': '_',
                    '[': '_',
                    ']': '_',
                    '(': '_',
                    ')': '_',
                    '>': '_',
                    '<': '_',
                    '#': '_',
                    '+': '_',
                    '&': '_',
                    '$': '_',
                    "'": '',
                    '"': '', }

    attr_name = in_name.strip()

    for to_replace, replacement in replace_dict.items():
        if to_replace in attr_name:
            for amount in reversed(range(1, 10)):
                if to_replace * amount in attr_name:

                    attr_name = attr_name.lstrip(to_replace * amount).rstrip(to_replace * amount).replace(to_replace * amount, replacement)
    return attr_name.casefold()


def filename_to_attr_name(in_file, keep_ext=False):
    attr_name = in_file
    if os.path.sep in attr_name or '/' in attr_name:
        attr_name = os.path.basename(attr_name)
    if keep_ext is False:
        attr_name = os.path.splitext(attr_name)[0]
    return to_attr_name(attr_name)
# region[Main_Exec]


if __name__ == '__main__':
    pass

# endregion[Main_Exec]



import os
import shutil
import click
from dotenv import load_dotenv, find_dotenv
from gidappdata.utility.functions import pathmaker, create_folder, create_file, loadjson, appendwriteit
from gidappdata.utility.extended_dotenv import find_dotenv_everywhere
from gidappdata.cli.skeleton_tree import serialize_all_prebuilts, DirSkeletonReader, SkeletonInstructionItem, get_all_prebuilts
from pprint import pprint
from gidappdata.cli.tree_render import print_tree
from functools import partial
from textwrap import dedent


THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))

SERIALIZED_DIR = pathmaker(DirSkeletonReader.serialized_prebuilts_folder)


def spec_appendwriteit(filepath, data):
    with open(filepath, 'a') as f:
        f.write(data + '\n')


def get_all_serialized_skeletons():
    serialize_all_prebuilts()
    _out = {}
    for serialized_skeleton in os.scandir(SERIALIZED_DIR):
        if os.path.isfile(serialized_skeleton) and serialized_skeleton.name.endswith('.json'):
            _out[serialized_skeleton.name.replace('.json', '')] = loadjson(serialized_skeleton.path)
    return _out


def create_user_data_setup(path):
    user_data_setup_file = pathmaker(path, 'user_data_setup.py')
    with open(user_data_setup_file, 'w') as uds_file:
        uds_file.write(dedent("""from gidappdata import SupportKeeper
                            from gidappdata.utility.extended_dotenv import find_dotenv_everywhere
                            import os
                            import dotenv
                            from .bin_data import bin_archive_data
                            dotenv.load_dotenv(find_dotenv_everywhere('project_meta_data.env'))

                            THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))
                            DATA_DIR = os.path.join(THIS_FILE_DIR, 'data_pack')
                            CONSTRUCTION_INFO_FILE = os.path.join(THIS_FILE_DIR, 'construction_info.env')


                            if os.path.isfile(CONSTRUCTION_INFO_FILE):
                                dotenv.load_dotenv(CONSTRUCTION_INFO_FILE)

                            if os.path.isfile('dev.trigger') is True:
                                SupportKeeper.set_dev(True, DATA_DIR)
                            SupportKeeper.set_archive_data(bin_archive_data)
                            """))


def create_dev_env_trigger(in_path):
    _path = pathmaker(in_path, 'dev.trigger')
    create_file(_path, '')


def select_skeleton(skeleton_selection, skeleton_category=None):
    category = 'standard' if skeleton_category is None else skeleton_category.casefold()
    selection = skeleton_selection.casefold()
    json_data = None
    all_serialized_skeletons = get_all_serialized_skeletons()
    for serialized_name in all_serialized_skeletons:
        if serialized_name.casefold() == f"[{category}]_{selection}":
            json_data = all_serialized_skeletons[serialized_name]
    if json_data is None:
        raise KeyError(f"unable to find serialized data for selection '{skeleton_selection}' and category '{category}'")
    return SkeletonInstructionItem.from_dict(json_data)


def build_target_skeleton(target_dir, skeleton_selection, skeleton_category, overwrite):
    path = pathmaker(target_dir, 'init_userdata')
    skeleton_tree = select_skeleton(skeleton_selection, skeleton_category)
    skeleton_tree.set_root_path(pathmaker(path, skeleton_tree.name))
    skeleton_tree.start_build(overwrite=overwrite)
    create_file(pathmaker(path, '__init__.py'), '')
    create_dev_env_trigger(path)
    create_user_data_setup(path)


@click.group()
def cli():
    pass


@cli.command(name='build')
@click.argument('target_dir')
@click.argument('skeleton_selection')
@click.option('-c', '--skeleton-category', default=None)
@click.option('--overwrite/--no-overwrite', '-o/-no', default=False)
def to_build_target_skeleton(target_dir, skeleton_selection, skeleton_category, overwrite):
    build_target_skeleton(target_dir, skeleton_selection, skeleton_category, overwrite)


@cli.command(name='list')
@click.option('--tree/--no-tree', '-t/-nt', default=True)
@click.option('-t', '--to-file', default=None)
def list_available(tree, to_file):
    all_prebuilts = get_all_prebuilts()
    output_func = print if to_file is None else partial(spec_appendwriteit, to_file)
    for category, value in all_prebuilts.items():
        category = category.replace('prebuilt_', '')
        for name, path in value.items():

            output_func('\n\n#################################\n')
            output_func('Category: ' + category.upper() + ' --> ' + 'Name: ' + name)
            if tree:
                path = os.path.dirname(path)
                print_tree(path, category, output_function=output_func)
            output_func('\n#################################\n\n')


if __name__ == '__main__':
    cli()



import os
import sys

os.chdir(sys.argv[1])
PROJECT_NAME = sys.argv[2]
REL_ACTIVATE_SCRIPT_PATH = './.venv/Scripts/activate.bat'
REPLACEMENT = r"""@echo off

set FILEFOLDER=%~dp0

pushd %FILEFOLDER%
rem ----------------------------------------------------------------
cd ..\..\tools
echo ##################### setting vars from %cd%\_project_meta.env
for /f %%i in (_project_meta.env) do set %%i
rem ----------------------------------------------------------------
popd
"""


def create_project_meta_env_file():
    os.chdir('../')
    _workspacedirbatch = os.getcwd()
    _toplevelmodule = os.path.join(_workspacedirbatch, PROJECT_NAME)
    _main_script_file = os.path.join(_toplevelmodule, '__main__.py')
    with open("tools/_project_meta.env", 'w') as envfile:
        envfile.write(f'WORKSPACEDIR={_workspacedirbatch}\n')
        envfile.write(f'TOPLEVELMODULE={_toplevelmodule}\n')
        envfile.write(f'MAIN_SCRIPT_FILE={_main_script_file}\n')
        envfile.write(f'PROJECT_NAME={PROJECT_NAME}\n')


def modify_activate_bat():

    with open(REL_ACTIVATE_SCRIPT_PATH, 'r') as origbat:
        _content = origbat.read()
    if REPLACEMENT not in _content:
        _new_content = _content.replace(r'@echo off', REPLACEMENT)
        with open(REL_ACTIVATE_SCRIPT_PATH, 'w') as newbat:
            newbat.write(_new_content)


if __name__ == '__main__':
    create_project_meta_env_file()
    modify_activate_bat()



# region [Imports]

# * Standard Library Imports -->
import gc
import os
import re
import sys
import json
import lzma
import time
import queue
import logging
import platform
import subprocess
from enum import Enum, Flag, auto
from time import sleep
from pprint import pprint, pformat
from typing import Union
from datetime import tzinfo, datetime, timezone, timedelta
from functools import wraps, lru_cache, singledispatch, total_ordering, partial
from contextlib import contextmanager
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from multiprocessing import Pool
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

# * Third Party Imports -->
# import requests
# import pyperclip
# import matplotlib.pyplot as plt
# from bs4 import BeautifulSoup
# from dotenv import load_dotenv
# from github import Github, GithubException
# from jinja2 import BaseLoader, Environment
# from natsort import natsorted
# from fuzzywuzzy import fuzz, process

# * Gid Imports -->
import gidlogger as glog

from gidappdata.utility import readit, readbin, writebin, writeit, writejson, loadjson, linereadit, pathmaker, create_folder, create_file, clearit, pickleit, get_pickled, read_file
import base64

# endregion[Imports]


# region [AppUserData]

# endregion [AppUserData]

# region [Logging]

log = logging.getLogger('gidappdata')
log.info(glog.imported(__name__))

# endregion[Logging]

THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))


def json_parts(item):
    if item.typus is not SkeletonTypus.Root:
        yield (item.parent.name, item.name, item.typus.value, item.content)

    if item.typus is not SkeletonTypus.File:
        for child in item.children:
            yield from json_parts(child)


def json_it(item, _file):
    _out = {}
    for _parent, _name, _typus, _content in json_parts(item):
        if _parent not in _out:
            _out[_parent] = []
        if isinstance(_content, bytes):
            _content = base64.b64encode(_content)
        _out[_parent].append((_name, _typus, _content))
    writejson(_out, _file, indent=2, sort_keys=False)


class SkeletonTypus(Enum):
    Folder = 'folder'
    File = 'file'
    Root = 'root'


class SkeletonInstructionBaseItem:
    pass


class SkeletonInstructionItem(SkeletonInstructionBaseItem):
    serialize_strategies = {'pickle': (pickleit, '.pkl'), 'json': (json_it, '.json')}
    standard_root_name = 'data_pack'

    def __init__(self, name: str, typus: SkeletonTypus, parent: SkeletonInstructionBaseItem = None, content=None):
        self.name = name
        self.typus = typus
        self.parent = parent
        self.children = [] if self.typus in [SkeletonTypus.Folder, SkeletonTypus.Root] else None
        self.content = content

    @staticmethod
    def _make_child_keyname(child):
        return child.name.replace('.', '__')

    def add_child_item(self, new_child: SkeletonInstructionBaseItem):
        if self.typus is SkeletonTypus.File:
            raise AttributeError("Files can not have children Items")
        if any(new_child.name.casefold() == existing_child.name.casefold() for existing_child in self.children):
            raise FileExistsError(f"The {new_child.typus.value} {new_child.name} already exist inside this Folder")
        self.children.append(new_child)
        new_child.parent = self

    @property
    def path(self):
        if self.parent is None or self.typus is SkeletonTypus.Root:
            return pathmaker(self.name)
        else:
            return pathmaker(self.parent.path, self.name)

    def set_root_path(self, root_path):
        if self.typus is SkeletonTypus.Root:
            self.name = root_path
        elif self.parent is None:
            raise AttributeError("This method can not be used on items that are not linked utimately to an ROOT Item")
        else:
            self.parent.set_root_path(root_path)

    def get_paths(self):
        yield self.path
        if self.typus is not SkeletonTypus.File:
            for child in self.children:
                yield from child.get_paths()

    def _build(self, overwrite=False):
        if self.typus is SkeletonTypus.Root:
            create_folder(self.path, False)
        elif self.typus is SkeletonTypus.Folder:
            create_folder(self.path, overwrite=overwrite)

        elif self.typus is SkeletonTypus.File:
            content = '' if self.content is None else self.content
            create_file(self.path, content, overwrite=overwrite)

        if self.children is not None:
            for child in self.children:
                child._build(overwrite=overwrite)

    def start_build(self, overwrite=False, from_top=True):
        if self.typus is not SkeletonTypus.Root and from_top is True:
            if self.parent is not None:
                self.parent.start_build(overwrite, from_top)
            else:
                raise AttributeError("This method with the argument 'from_top'=True can not be used on items that are not linked utimately to an ROOT Item")
        else:
            self._build(overwrite)

    def serialize(self, save_file_path, strategy='json', from_top=True):
        if self.typus is not SkeletonTypus.Root and from_top is True:
            if self.parent is not None:
                self.parent.serialize(save_file_path, strategy, from_top)
            else:
                raise AttributeError("This method with the argument 'from_top'=True can not be used on items that are not linked utimately to an ROOT Item")
        elif self.typus is not SkeletonTypus.Root and from_top is False:
            self.parent = None
            _func, _extension = self.serialize_strategies[strategy]
            _file = save_file_path + _extension if '.' not in os.path.basename(save_file_path) else save_file_path
            _func(self, _file)
        elif self.typus is SkeletonTypus.Root:
            _func, _extension = self.serialize_strategies[strategy]
            _file = save_file_path + _extension if '.' not in os.path.basename(save_file_path) else save_file_path
            _func(self, _file)

    @classmethod
    def from_json_file(cls, json_file_path):
        data = loadjson(json_file_path)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, dict_data):
        root = cls(name=cls.standard_root_name, typus=SkeletonTypus.Root)
        for parent, value in dict_data.items():
            for name, typus, content in value:
                if parent == cls.standard_root_name:
                    root.add_child_item(SkeletonInstructionItem(name=name, typus=SkeletonTypus(typus), content=content))
                else:
                    root.find_node(parent).add_child_item(SkeletonInstructionItem(name=name, typus=SkeletonTypus(typus), content=content))
        return root

    def __getattr__(self, name):
        _out = None
        for child in self.children:
            if self._make_child_keyname(child).casefold() == name.casefold():
                _out = child
        if _out is None:
            raise AttributeError(name)
        return _out

    def all_nodes(self):
        if self.parent is not None:
            self.parent.all_nodes()
        else:
            yield self.name, self
            yield from self._all_nodes_walker()

    def _all_nodes_walker(self):
        if self.typus is not SkeletonTypus.File:
            for child_item in self.children:
                yield child_item.name, child_item
                yield from child_item._all_nodes_walker()

    def find_node(self, name):
        for node_name, node_object in self.all_nodes():
            if node_name.casefold() == name.casefold():
                return node_object


class DirSkeletonReader:
    exclude = {'__pycache__', '.git'}
    exclude = set(map(lambda x: x.casefold(), exclude))
    prebuilt_basefolder = pathmaker(THIS_FILE_DIR, '../data/skeletons')
    prebuilt_folders = loadjson(pathmaker(THIS_FILE_DIR, 'prebuilt_folders.json'))
    serialized_prebuilts_folder = pathmaker(THIS_FILE_DIR, '../data/serialized_skeletons')

    def __init__(self, start_folder_path):
        self.start_folder_path = start_folder_path
        self.start_folder_name = os.path.basename(self.start_folder_path)
        self.skeleton_tree = None
        self._make_skeleton_tree()

    def _make_skeleton_tree(self):
        self.skeleton_tree = SkeletonInstructionItem(self.start_folder_name, SkeletonTypus.Root)
        for dirname, folderlist, filelist in os.walk(self.start_folder_path, followlinks=False):
            if all(ex_item.casefold() not in dirname.casefold() for ex_item in self.exclude):
                parent = os.path.basename(dirname)
                for foldername in folderlist:
                    if foldername.casefold() not in self.exclude:
                        self.skeleton_tree.find_node(parent).add_child_item(SkeletonInstructionItem(foldername, SkeletonTypus.Folder))
                for filename in filelist:
                    if filename.casefold() not in self.exclude:
                        self.skeleton_tree.find_node(parent).add_child_item(SkeletonInstructionItem(filename, SkeletonTypus.File, content=read_file(pathmaker(dirname, filename))))

    def rename_node(self, node_name, new_name):
        node = self.find_node(node_name)
        node.name = new_name

    def __getattr__(self, name):
        return getattr(self.skeleton_tree, name)


def quick_check_skeleton():
    item_dict = {
        'first_folder': [('first_subfolder', SkeletonTypus.Folder, None),
                         ('second_subfolder', SkeletonTypus.Folder, None),
                         ('first_folder_image_file.jpg', SkeletonTypus.File, readbin(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppData\tests\skeleton_tree_test\afa_logoover_ca.jpg")),
                         ('first_folder_ini_file.ini', SkeletonTypus.File, readit(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppData\tests\skeleton_tree_test\example_cfg.ini"))],
        'second_folder': [('something_file.txt', SkeletonTypus.File, 'this')]}
    root = SkeletonInstructionItem(name='root', typus=SkeletonTypus.Root)
    root.add_child_item(SkeletonInstructionItem(name='first_folder', typus=SkeletonTypus.Folder))
    root.add_child_item(SkeletonInstructionItem(name='second_folder', typus=SkeletonTypus.Folder))
    root.add_child_item(SkeletonInstructionItem(name='text_file.txt', typus=SkeletonTypus.File, content='this is a test text'))

    for key, value in item_dict.items():
        for _name, _typus, _content in value:
            root.children[key].add_child_item(SkeletonInstructionItem(name=_name, typus=_typus, content=_content))
    root.first_folder.first_subfolder.add_child_item(SkeletonInstructionItem(name='nested_test.json', typus=SkeletonTypus.File, content='{}'))
    print(root.find_node('first_folder_ini_file.ini').name)


def get_all_prebuilts():
    skeleton_selections = {}
    for prebuilt_typus in DirSkeletonReader.prebuilt_folders:
        src_folder = pathmaker(DirSkeletonReader.prebuilt_basefolder, prebuilt_typus)
        if src_folder not in skeleton_selections:
            skeleton_selections[prebuilt_typus] = {}
        for skeleton in os.scandir(src_folder):
            if os.path.isdir(skeleton.path):
                skeleton_selections[prebuilt_typus][skeleton.name] = pathmaker(skeleton.path, 'data_pack')
    return skeleton_selections


def serialize_all_prebuilts():
    for file in os.scandir(DirSkeletonReader.serialized_prebuilts_folder):
        if os.path.isfile(file):
            os.remove(file.path)
    for typus_folder, value in get_all_prebuilts().items():
        typus = typus_folder.replace('prebuilt_', '')
        for name, path in value.items():
            name = f"[{typus}]_{name}"
            skeleton_tree = DirSkeletonReader(path)
            skeleton_tree.serialize(pathmaker(DirSkeletonReader.serialized_prebuilts_folder, name))


# region[Main_Exec]


if __name__ == '__main__':
    pass
# endregion[Main_Exec]



from pathlib import Path
from typing import Union
import sys
# prefix components:
space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '
headline_high_start = '┌'
headline_low_start = '└'
headline_body = '─'
headline_high_end = '┐'
headline_low_end = '┘'
headline_sides = '│'


def tree(dir_path: Path, prefix: str = '    '):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """

    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(path, prefix=prefix + extension)


def create_tree(dir_path: Path, prefix: str = '    ', category=None):
    name = 'Name: ' + dir_path.name
    if category is not None:
        category = 'Category: ' + category.upper()
        length_name = max([len(name), len(category)])
        length_dif = length_name - min([len(name), len(category)])
    else:
        length_name = len(name)
    yield headline_high_start + headline_body * length_name + headline_high_end
    if category is not None and len(name) < len(category):
        yield headline_sides + name + ' ' * length_dif + headline_sides
    else:
        yield headline_sides + name + headline_sides
    if category is not None:
        if len(category) < len(name):
            yield headline_sides + category + ' ' * length_dif + headline_sides
        else:
            yield headline_sides + category + headline_sides
    yield headline_low_start + headline_body * length_name + headline_low_end
    yield from tree(dir_path, prefix)


def print_tree(dir_path: Union[Path, str], category=None, output_function=print):
    if isinstance(dir_path, str):
        dir_path = Path(dir_path)
    for line in create_tree(dir_path, '    ', category):
        output_function(line)


if __name__ == '__main__':
    print_tree(sys.argv[1])



import os
from dotenv import load_dotenv, find_dotenv


def _is_appdata_object(in_object):
    try:
        _name = in_object.__class__.__name__
        return _name == 'AppDataStorager'
    except AttributeError:
        return False


def find_dotenv_everywhere(filename: str = '.env', start_dir=None, lower_folder: bool = True, upper_folder=True, raise_error_if_not_found: bool = False):
    start_dir = os.getcwd() if start_dir is None else start_dir
    start_dir = str(start_dir) if _is_appdata_object(start_dir) is True else start_dir
    dotfile = None
    if lower_folder:
        for dirname, _, filelist in os.walk(start_dir):
            for file in filelist:
                if file == filename:
                    print()
                    dotfile = os.path.join(dirname, file)
    if dotfile is None and upper_folder:
        _old_dir = os.getcwd()
        os.chdir(start_dir)
        dotfile = find_dotenv(filename, raise_error_if_not_found, usecwd=True)

        os.chdir(_old_dir)
    if dotfile is None:
        dotfile = ''
    return dotfile



# region [Imports]


import lzma
import os
from dotenv import load_dotenv
import zipfile
import os
import base64
import logging
import gidlogger as glog
from gidconfig.standard import ConfigHandler

from gidappdata.standard_appdata.appdata_storager import AppDataStorager
from gidappdata.utility.functions import pathmaker, to_attr_name, filename_to_attr_name, create_folder, create_file, readit, writeit
from gidappdata.utility.extended_dotenv import find_dotenv_everywhere
from gidappdata.utility.exceptions import ConstructionEnvDataMissing, DevSettingError
from gidappdata.cli.pack_and_bin_and_py_data import generate_user_data_binfile
# endregion [Imports]


# region [Logging]

log = logging.getLogger('gidappdata')
log.info(glog.imported(__name__))

# endregion [Logging]


class ParaStorageKeeperMetaHelper(type):
    def __getattr__(cls, name):
        _out = ParaStorageKeeper.configs.get(name, None)
        if _out is None:
            _out = ParaStorageKeeper.app_info.get(name)
        if _out is None:
            raise AttributeError
        return _out


class ParaStorageKeeper(metaclass=ParaStorageKeeperMetaHelper):
    # region [ClassAttributes]

    is_init = False
    appdata = None
    configs = {}
    construction_env_filename = 'construction_info.env'
    app_info = {'app_name': None, 'author_name': None, 'uses_base64': None, 'clean': True, 'dev': False, 'redirect': '', 'log_folder': '', "is_unpacked": False}
    config_handler = ConfigHandler
    archive_data = None
    # endregion[ClassAttributes]

    @staticmethod
    def _unzip(root_dir, zip_file, overwrite: bool = False):
        # sourcery skip: simplify-boolean-comparison
        with zipfile.ZipFile(zip_file, 'r') as zipf:
            for item in zipf.namelist():
                _info = zipf.getinfo(item)
                if _info.is_dir() is True:
                    create_folder(pathmaker(root_dir, item))
                else:
                    if os.path.isfile(pathmaker(root_dir, item)) is False or overwrite is True:
                        zipf.extract(item, pathmaker(root_dir))
                        log.debug("extracted file '%s' because it didn't exist", pathmaker(root_dir, item))
                    else:
                        log.debug("file '%s' is already existing and overwrite is 'False' so file was not extracted", pathmaker(root_dir, item))
        log.info('unzipping finished')

    @classmethod
    def set_experimental_confighandler(cls):
        from gidconfig.experimental import GidAttConfigIni
        cls.config_handler = GidAttConfigIni

    @classmethod
    def set_clean(cls, setting: bool):
        cls.app_info['clean'] = setting

    @classmethod
    def set_dev(cls, setting: bool, redirect=None, log_folder=None):
        # sourcery skip: simplify-boolean-comparison
        cls.app_info['dev'] = setting
        if setting is True:
            if redirect is None:
                raise DevSettingError()
            cls.app_info['redirect'] = pathmaker(redirect)
            cls.app_info['log_folder'] = pathmaker(log_folder)

    @classmethod
    def set_archive_data(cls, archive_data: bytes):
        cls.archive_data = archive_data

    @staticmethod
    def checked_get_env(env_var_name):
        _out = os.getenv(env_var_name)
        if _out is None:
            raise ConstructionEnvDataMissing(env_var_name)
        if _out.casefold() in ['true', 'yes', '1']:
            _out = True
        elif _out.casefold() in ['false', 'no', '0']:
            _out = False
        else:
            _out = _out
        return _out

    @classmethod
    def _archive_from_bin(cls, bin_data, name: str = 'user_data_archive', ext: str = 'zip', uses_base64: bool = False):
        _file = pathmaker(str(cls.appdata), name + '.' + ext)
        with open(_file, 'wb') as archfile:
            _bin_data = bin_data if not uses_base64 else base64.b64decode(bin_data)
            archfile.write(_bin_data)
        return _file

    @classmethod
    def unpack_archive(cls, in_archive, clean: bool, uses_base64: bool):
        _file = cls._archive_from_bin(in_archive, uses_base64=uses_base64)
        cls._unzip(str(cls.appdata), _file, False)
        if clean:
            os.remove(_file)

    @classmethod
    def find_construct_env(cls):
        for dirname, folderlist, filelist in os.walk(os.getcwd()):
            for file in filelist:
                if file == cls.construction_env_filename:
                    return pathmaker(dirname, file)

    @classmethod
    def set_unpacked(cls):
        cls.app_info['is_unpacked'] = True

    @classmethod
    def initialize(cls, archive_data=None):
        if cls.is_init is True:
            return
        load_dotenv(find_dotenv_everywhere(cls.construction_env_filename))
        for info in cls.app_info:
            if cls.app_info[info] is None:
                cls.app_info[info] = cls.checked_get_env(info.upper())
        redirect = None if cls.app_info['redirect'] == '' else cls.app_info['redirect']
        log_folder = None if cls.app_info['log_folder'] == '' else cls.app_info['log_folder']
        archive_data = cls.archive_data if archive_data is None else archive_data
        cls.appdata = AppDataStorager(cls.app_info['author_name'], cls.app_info['app_name'], cls.app_info['dev'], redirect, log_folder)

        if cls.app_info['dev'] is False or cls.app_info.get('is_unpacked') is False:
            cls.unpack_archive(archive_data, cls.app_info['clean'], cls.app_info['uses_base64'])
            cls.set_unpacked()
        if os.path.isdir(cls.appdata['config']) is True:
            for file in os.scandir(cls.appdata['config']):
                if file.name.endswith('.ini') and 'config' in file.name:
                    name = filename_to_attr_name(file.name)
                    cls.configs[name] = ConfigHandler(cls.appdata[file.name])
        cls.is_init = True

    @classmethod
    def get_appdata(cls):
        if cls.is_init is False:
            cls.initialize()
        return cls.appdata

    @classmethod
    def get_config(cls, config_name):
        if cls.is_init is False:
            cls.initialize()
        return cls.configs.get(config_name)


if __name__ == '__main__':
    pass



import pytest
import os
import shutil
from gidappdata import SupportKeeper, AppdataProvider, ConfigHandler
from gidappdata.utility import pathmaker, loadjson, writejson, writeit, readit, readbin, linereadit, appendwriteit


def test_appinfo(deployed_supportkeeper):
    assert SupportKeeper.dev == False
    assert SupportKeeper.author_name == "BrocaProgs"
    assert SupportKeeper.app_name == "Test_App"
    assert SupportKeeper.uses_base64 == True
    assert str(SupportKeeper.appdata) == deployed_supportkeeper[1]


def test_config_detection(deployed_supportkeeper):
    configs = ['solid_config', 'user_config', 'db_config']
    for config in configs:
        assert config in SupportKeeper.configs
    assert isinstance(SupportKeeper.solid_config, ConfigHandler)
    assert isinstance(SupportKeeper.user_config, ConfigHandler)
    assert isinstance(SupportKeeper.db_config, ConfigHandler)



"""
GidAppData provides an Userdata deployment, with appdirs as backendend, but easy access to files saved there. Also provides an option to 'fake' deploy it while still developing
"""

__version__ = '0.1.8'


from gidappdata.standard_appdata import *


# log = logging.getLogger('gidappdata')



import pytest
from gidappdata.cli.skeleton_tree import SkeletonInstructionItem, SkeletonTypus
from tempfile import TemporaryDirectory
from gidappdata.utility.functions import readbin, readit, pathmaker
import os

THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture
def temp_root_dir():
    with TemporaryDirectory() as tempdir:
        yield tempdir
    print('closed temp dir')


@pytest.fixture
def simple_directory_tree():

    item_dict = {
        'first_folder': [('first_subfolder', SkeletonTypus.Folder, None),
                         ('second_subfolder', SkeletonTypus.Folder, None),
                         ('first_folder_image_file.jpg', SkeletonTypus.File, readbin(pathmaker('tests', 'skeleton_tree_test', 'afa_logoover_ca.jpg'))),
                         ('first_folder_ini_file.ini', SkeletonTypus.File, readit(pathmaker('tests', 'skeleton_tree_test', 'example_cfg.ini')))],
        'second_folder': [('something_file.txt', SkeletonTypus.File, 'this')]}
    root = SkeletonInstructionItem(name='root', typus=SkeletonTypus.Root)
    root.add_child_item(SkeletonInstructionItem(name='first_folder', typus=SkeletonTypus.Folder))
    root.add_child_item(SkeletonInstructionItem(name='second_folder', typus=SkeletonTypus.Folder))
    root.add_child_item(SkeletonInstructionItem(name='text_file.txt', typus=SkeletonTypus.File, content='this is a test text'))

    for key, value in item_dict.items():
        for _name, _typus, _content in value:
            getattr(root, key).add_child_item(SkeletonInstructionItem(name=_name, typus=_typus, content=_content))
    root.first_folder.first_subfolder.add_child_item(SkeletonInstructionItem(name='nested_test.json', typus=SkeletonTypus.File, content='{}'))
    yield root


@pytest.fixture
def access_test_tree():
    root = SkeletonInstructionItem(name='root', typus=SkeletonTypus.Root)
    first_folder_item = SkeletonInstructionItem(name='first_folder', typus=SkeletonTypus.Folder)
    second_folder_item = SkeletonInstructionItem(name='second_folder', typus=SkeletonTypus.Folder)
    first_sub_folder_item = SkeletonInstructionItem(name='first_sub_folder', typus=SkeletonTypus.Folder)
    sub_folder_file_item = SkeletonInstructionItem(name='sub_folder_file.txt', typus=SkeletonTypus.File, content='this is a subfolder txt file')
    root.add_child_item(first_folder_item)
    root.add_child_item(second_folder_item)
    first_folder_item.add_child_item(first_sub_folder_item)
    first_sub_folder_item.add_child_item(sub_folder_file_item)
    yield root, first_folder_item, second_folder_item, first_sub_folder_item, sub_folder_file_item



import pytest
from gidappdata import AppDataStorager, SupportKeeper
import os
import shutil
from gidappdata.utility.functions import readit, writeit, writebin, create_folder, writejson, pickleit, pathmaker
from .bin_data import bin_archive_data

THIS_FILE_DIR = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture
def simple_appdata_storage():
    appdata = AppDataStorager('test_author', 'test_app_name')
    yield appdata
    try:
        appdata.clean(appdata.AllFolder)
    except FileNotFoundError as error:
        print(str(error))


@pytest.fixture
def filled_appdata_storage(simple_appdata_storage):
    _first_folder = pathmaker(str(simple_appdata_storage), 'first_folder')
    create_folder(_first_folder)
    create_folder(pathmaker(_first_folder, 'data_subfolder'))
    create_folder(pathmaker(_first_folder, 'image_subfolder'))
    writeit(pathmaker(_first_folder, 'data_subfolder', 'test_1.txt'), 'this is a test text file!')
    writeit(pathmaker(_first_folder, 'data_subfolder', 'test_2.txt'), 'this is another test text file!')
    writejson({}, pathmaker(_first_folder, 'data_subfolder', 'test_dict.json'))
    source_folder = pathmaker('tests', 'standard_appdata_test')
    target_folder = pathmaker(_first_folder, 'image_subfolder')
    images = ['test_image.ico', 'test_image.jpg', 'test_image.png']
    for image in images:
        shutil.copyfile(pathmaker(source_folder, image), pathmaker(target_folder, image))
    yield simple_appdata_storage


@pytest.fixture(scope="session")
def construction_env():
    path = pathmaker(THIS_FILE_DIR, 'construction_info.env')
    author_name = "BrocaProgs"
    app_name = "Test_App"
    with open(path, 'w') as const_file:
        const_file.write("USES_BASE64=True\n")
        const_file.write(f"AUTHOR_NAME={author_name}\n")
        const_file.write(f"APP_NAME={app_name}")
    yield author_name, app_name
    os.remove(path)


@pytest.fixture(scope="session")
def deployed_supportkeeper(construction_env):
    save_path = pathmaker(os.getenv('APPDATA'), construction_env[0], construction_env[1])
    SupportKeeper.set_archive_data(bin_archive_data)
    print(SupportKeeper.get_appdata())
    yield SupportKeeper.get_appdata(), save_path
    shutil.rmtree(os.path.dirname(save_path))
    assert os.path.isdir(os.path.dirname(save_path)) is False
